﻿using System;
using System.Threading;
using UnityEngine;

public interface IDestroyTokenWrapper : IDisposable
{
    CancellationToken CancellationToken { get; }
    void CancelActiveToken();
}

public static class DestroyTokenWrapperFactory
{
    public static IDestroyTokenWrapper CreateIDestroyTokenWrapper(MonoBehaviour monoBehaviour)
    {
#if UNITY_2022_2_OR_NEWER
        return new DestroyToken2022OrLater(monoBehaviour);
#else
        return new DestroyTokenWrapperBefore2022();
#endif
    }
}

#if UNITY_2022_2_OR_NEWER

internal class DestroyToken2022OrLater : IDestroyTokenWrapper
{
    private readonly MonoBehaviour m_Monobehaviour;
    private CancellationTokenSource m_InternalCts;
    private CancellationTokenSource m_LinkedCts;
    private bool m_isDisposed = false;
    
    public DestroyToken2022OrLater(MonoBehaviour monobehaviour)
    {
        m_Monobehaviour = monobehaviour;
        m_InternalCts = new CancellationTokenSource();
        m_LinkedCts = CancellationTokenSource.CreateLinkedTokenSource(m_Monobehaviour.destroyCancellationToken, m_InternalCts.Token);
#if UNITY_EDITOR
        UnityEditor.EditorApplication.wantsToQuit += WantsToQuitCallback;
#endif
    }

    private bool WantsToQuitCallback()
    {
#if UNITY_EDITOR
        UnityEditor.EditorApplication.wantsToQuit -= WantsToQuitCallback;
#endif
        Dispose();
        return true;
    }

    public void Dispose()
    {
        if (m_isDisposed) return;
#if UNITY_EDITOR
        UnityEditor.EditorApplication.wantsToQuit -= WantsToQuitCallback;
#endif
        m_InternalCts.Cancel();
        m_InternalCts.Dispose();
        
        m_LinkedCts.Cancel();
        m_LinkedCts.Dispose();
        m_isDisposed = true;
    }
    public void CancelActiveToken()
    {
        if (m_isDisposed) return;
        m_InternalCts.Cancel();
        m_InternalCts.Dispose();
        m_LinkedCts.Cancel();
        m_LinkedCts.Dispose();
        
        m_InternalCts = new CancellationTokenSource();
        m_LinkedCts = CancellationTokenSource.CreateLinkedTokenSource(m_Monobehaviour.destroyCancellationToken, m_InternalCts.Token);
    }

    public CancellationToken CancellationToken
    {
        get
        {
            if (m_Monobehaviour == null || m_isDisposed) //not sure about any guarantees of when these operations happen - token, triggered, monobehaviour native handle nulled in C#, and m_Monobehaviour.destroyCancellationToken cancelled
            {
                return new CancellationToken(canceled: true);
            }
            return m_LinkedCts.Token;
        }
    }
}
#else
    internal class DestroyTokenWrapperBefore2022 : IDestroyTokenWrapper
    {
        private CancellationTokenSource cts = new CancellationTokenSource();
        private bool m_isDisposed = false;
        
        public void Dispose()
        {
            if (m_isDisposed) return;
            cts.Cancel();
            cts.Dispose();
            m_isDisposed = true;
        }

        public CancellationToken CancellationToken { get; }

        public void CancelActiveToken()
        {
            if (m_isDisposed) return;
            cts.Cancel();
            cts.Dispose();
            cts = new CancellationTokenSource();
        }
    }
#endif