﻿using System;
using System.Collections.Generic;
using Sos;
using UnityEngine;

[DefaultExecutionOrder(-50)]
public class StreamController : MonoBehaviour
{
    [Serializable]
    public class StreamControllerMonobehaviourConfiguration
    {
        [Tooltip("Set this to the camera that you will use to move around the scene. Geometry is updated based on the position of the camera.")]
        public Camera viewerCamera;

        [Space(10)]
        [Tooltip("Dynamically Occlusion Culling")]
        public bool occlusionCulling = true;
        public int modelLayer = -1;
        [Tooltip("Determine how many frames should be accumulated to decide whether an object should be hidden.")]
        [SerializeField]
        public int timeSlicing = 120;
        [Tooltip("How many rays are generated per frame.")]
        [SerializeField]
        public int raysPerFrame = 256;
        [SerializeField]
        public LayerMask raycastMask = 0;

        public void CopyTo(StreamControllerMonobehaviourConfiguration copy)
        {
            copy.viewerCamera = viewerCamera;
            occlusionCulling = copy.occlusionCulling;
            timeSlicing = copy.timeSlicing;
            raysPerFrame = copy.raysPerFrame;
            raycastMask = copy.raycastMask;
        }
    }
    public StreamControllerMonobehaviourConfiguration streamControllerMonobehaviourConfiguration = new StreamControllerMonobehaviourConfiguration();
    public StreamControllerImplementation.StreamControllerInitOptions streamControllerInitOptions = new StreamControllerImplementation.StreamControllerInitOptions();
    private IDestroyTokenWrapper m_destroyTokenWrapper;

    public void AddModel(StreamingModel model)
    {
        m_models.Add(model.m_implementation);

        OnAddStreamingModel?.Invoke(model);
    }

    public void RemoveModel(StreamingModel model)
    {
        m_models.Remove(model.m_implementation);

        OnRemoveStreamingModel?.Invoke(model);
    }

    private void Awake()
    {
        m_destroyTokenWrapper = DestroyTokenWrapperFactory.CreateIDestroyTokenWrapper(this);
    }

    private void Start()
    {
        m_implementation = new StreamControllerImplementation(streamControllerInitOptions);
        if (streamControllerMonobehaviourConfiguration.viewerCamera == null)
        {
            Debug.LogWarning("XrgManager's viewerCamera property must be set! Defaulting to Camera.main");
            streamControllerMonobehaviourConfiguration.viewerCamera = Camera.main;
        }

        int polygonStreamingLayer = LayerMask.NameToLayer("PolygonStreaming");
        streamControllerMonobehaviourConfiguration.modelLayer = polygonStreamingLayer;


        if (streamControllerMonobehaviourConfiguration.occlusionCulling && polygonStreamingLayer != -1)
        {
            int layerMask = streamControllerMonobehaviourConfiguration.raycastMask |= 1 << polygonStreamingLayer;
            m_occlusionCulling = new StreamingOcclusionCulling(streamControllerMonobehaviourConfiguration.timeSlicing, streamControllerMonobehaviourConfiguration.raysPerFrame, layerMask);
        }

    }

    private void Update()
    {
        if (!enabled) return;
        if (m_implementation != null)
            m_implementation.Update(m_models, streamControllerMonobehaviourConfiguration.viewerCamera, m_destroyTokenWrapper.CancellationToken);

        if (streamControllerMonobehaviourConfiguration.occlusionCulling)
            m_occlusionCulling?.Update(streamControllerMonobehaviourConfiguration.viewerCamera);
    }

    private void OnDisable()
    {
        m_destroyTokenWrapper?.CancelActiveToken();
        m_implementation?.OnDisable();
    }

    private void OnDestroy()
    {
        m_occlusionCulling?.OnDestroy();
        m_destroyTokenWrapper?.CancelActiveToken();
        m_destroyTokenWrapper?.Dispose();
        m_implementation?.OnDestroy();
    }

    private void OnValidate()
    {
        if (streamControllerMonobehaviourConfiguration.occlusionCulling &&
            streamControllerMonobehaviourConfiguration.viewerCamera == null)
        {
            //Debug.LogWarning("Occlusion Culling requires a viewer camera to be set, trying to set to main camera");
            if (Application.isPlaying == false)
                streamControllerMonobehaviourConfiguration.viewerCamera = Camera.main;
#if UNITY_EDITOR
            UnityEditor.EditorUtility.SetDirty(this);
#endif
        }
        if (Resources.Load<CacheSettings>("SosCacheSettings") == null)
        {
#if UNITY_EDITOR
            Debug.LogWarning("SosCacheSettings not found, creating default settings");

            var cacheSettings = ScriptableObject.CreateInstance<CacheSettings>();
            if (!UnityEditor.AssetDatabase.IsValidFolder("Assets/Resources"))
            {
                string folder = UnityEditor.AssetDatabase.CreateFolder("Assets", "Resources");
            }
            UnityEditor.AssetDatabase.CreateAsset(cacheSettings, "Assets/Resources/SosCacheSettings.asset");
            UnityEditor.EditorUtility.SetDirty(cacheSettings);
            UnityEditor.AssetDatabase.SaveAssetIfDirty(cacheSettings);
            UnityEditor.AssetDatabase.Refresh();
#endif
        }
    }

    private HashSet<ModelImplementation> m_models = new HashSet<ModelImplementation>();
    internal StreamControllerImplementation m_implementation;

    internal StreamingOcclusionCulling m_occlusionCulling;
    public Action<StreamingModel> OnAddStreamingModel { get; set; }
    public Action<StreamingModel> OnRemoveStreamingModel { get; set; }
}
