using System;
using System.Threading;
using System.Threading.Tasks;
using Sos;
using UnityEngine;


public class StreamingModel : MonoBehaviour
{
    public StreamController streamController;
    public string sourceUrl;
    public string hashCode;
    public Sos.NormalsMode normals = Sos.NormalsMode.OriginalVertexNormals;
    public bool lightProbeUsage = false;

    public bool customMaterials = false;
    public Material material;
    public Material transparentMaterial;

    public StreamingModelRuntimeOptions StreamingRuntimeOptionsSettings = new StreamingModelRuntimeOptions();
    [Range(0, 1)] private float alphaClipThreshold = 0.5f; // For now private/disabled

    private bool m_registered = false;
    public IDestroyTokenWrapper DestroyTokenWrapper { get; private set; }

    void SetupImpl()
    {
        if (DestroyTokenWrapper == null)
        {
            DestroyTokenWrapper = DestroyTokenWrapperFactory.CreateIDestroyTokenWrapper(this);
        }

        Sos.MaterialFlags materialFlags = new Sos.MaterialFlags();
        materialFlags.normals = normals;
        materialFlags.alphaClipThreshold = alphaClipThreshold;
        materialFlags.lightProbeUsage = lightProbeUsage;
        materialFlags.customMaterials = customMaterials;



        m_implementation = new Sos.ModelImplementation(sourceUrl, hashCode, material, transparentMaterial, materialFlags, StreamingRuntimeOptionsSettings, DestroyTokenWrapper.CancellationToken);
        m_implementation.GeometryTransform = transform;
        m_implementation.ViewerCamera = streamController.streamControllerMonobehaviourConfiguration.viewerCamera;
        m_implementation.UseOcclusionCulling = streamController.streamControllerMonobehaviourConfiguration.occlusionCulling;
        m_implementation.ModelLayer = streamController.streamControllerMonobehaviourConfiguration.modelLayer;
        m_implementation.OnLoadFail = m_OnFail;
        m_implementation.OnLoadSuccess = m_OnSuccess;
        //m_implementation.Quality = quality;
    }

    void Update()
    {
        if (streamController == null)
        {
            return;
        }
        if (m_implementation == null || m_implementation.IsDiscarded)
        {
            SetupImpl();
        }

        if (!m_registered)
        {
            streamController.AddModel(this);
            m_registered = true;
        }
    }
    private void OnDisable()
    {
        DestroyTokenWrapper?.CancelActiveToken(); //cancel downloads and other things, but this may also cancel initialization, so have to check for init again in enable
        if (m_registered && streamController != null)
        {
            streamController.RemoveModel(this);

            m_registered = false;
        }
        m_implementation?.Discard();
        m_implementation = null;
    }

    void OnDestroy()
    {
        DestroyTokenWrapper?.CancelActiveToken();
        DestroyTokenWrapper?.Dispose();
        if (streamController != null)
        {
            streamController.RemoveModel(this);
            m_registered = false;
        }
        if (m_implementation != null)
        {
            m_implementation.Discard();
            m_implementation = null;
        }
    }

    internal Sos.ModelImplementation m_implementation;
    private Action m_OnSuccess;
    private Action<Exception> m_OnFail;


    public void SetLoadedCallback(Action onSuccess)
    {
        if (m_implementation != null)
        {
            m_implementation.OnLoadSuccess = onSuccess;
        }
        else
        {
            m_OnSuccess = onSuccess;
        }
    }

    public void SetLoadFailedCallback(Action<Exception> onFail)
    {
        if (m_implementation != null)
        {
            m_implementation.OnLoadFail = onFail;
        }
        else
        {
            m_OnFail = onFail;
        }
    }

    public static Task<Mesh> FetchColliderMesh(string url)
    {
        return Sos.ModelImplementation.FetchColliderMesh(url);
    }
}
